from __future__ import annotations
from pathlib import Path


__all__ = ("extract_url_from_contents", "extract_url")


def _half(what: str, where: str) -> tuple[str, str]:
    ret: list[str] = what.split(where)
    assert len(ret) == 2, f"{what} does not contain {where} exactly once"
    return ret[0], ret[1]


def extract_url_from_contents(data: str, extension: str) -> str:
    """
    Extract the URL from the file data
    """
    upper: str
    lower: str
    if ".url" == extension:
        upper = _half(data, "URL=")[1]
    elif ".desktop" == extension:
        lower = _half(data, "URL=")[1]
        upper = _half(lower, "Icon=")[0]
    elif ".webloc" == extension:
        lower = _half(data, "<string>")[1]
        upper = _half(lower, "</string>")[0]
    else:
        raise ValueError(f"Unsupported extension: {extension}")
    url: str = upper.strip()
    assert len(url), "Failed to extract url"
    return url


def extract_url(file: Path):
    """
    Extract the URL from the file
    """
    file = file.resolve()
    assert file.exists(), f"{file} does not exist"
    assert file.suffix, f"{file} has no extension"
    with file.open("r") as f:
        try:
            data = f.read()
        except UnicodeDecodeError as e:
            raise RuntimeError(f"Failed to read: {file}") from e
    return extract_url_from_contents(data.strip(), file.suffix)



from functools import cache
from time import sleep

import requests

from manga.utils import split_on_num
from manga import sites

from .status import (
    Success,
    Tested,
    HasVol,
    NotInt,
    Tiny,
    Pattern,
    Exists,
    Missing,
    Broken,
    Unknown,
    BadRequest,
    PointFive,
)


_success = Success()


@cache
def _test_site(url: str) -> bool:
    return sites.test(url, timeout_retries=8)


def _test(left: str, right: str, x: float) -> bool:
    """
    Test if chapter x is found at a URL constructed from left, right, and x
    """
    if int(x) == x:
        return _test_site(f"{left}{int(x)}{right}")
    else:
        if _test_site(f"{left}{x}{right}"):
            return True
        return _test_site(f"{left}{str(x).replace('.', '-')}{right}")


# pylint: disable=too-many-return-statements,too-many-branches
def test_url(url: str) -> Tested:
    """
    If url is broken, return a failure class for it
    """
    if "vol" in url.lower():
        return HasVol()
    left, n, right = split_on_num(url)
    if n != int(n):
        return NotInt()
    elif n <= 5:
        return Tiny()
    elif "mangabuddy" in left and ("/mbx" in left or any(i.isalpha() for i in right)):
        return Pattern()
    test = lambda x: _test(left, right, x)
    try:
        if test(n) and not test(n - 1) and not test(5):
            return Exists()
        sleep(0.04)  # No DOS-ing
        if not test(n):
            if any(test(n + i) for i in (0.1, 0.5, 1, 1.1, 2, 2.1, 5, 10, 20)):
                return Missing()
            sleep(0.08)  # No DOS-ing
            if test(n - 0.5):
                return PointFive()
        sleep(0.01)  # No DOS-ing
        if not any(test(i) for i in (n, n - 1, 5, n + 0.1, n + 0.5, n - 0.5, n + 1, n + 1.1, n + 5)):
            return Broken()
        sleep(0.04)  # No DOS-ing
        return _success
    except sites.UnknownDomain:
        return Unknown()
    except requests.exceptions.RequestException as e:
        print(e)
        return BadRequest(e)



from __future__ import annotations
from typing import TYPE_CHECKING, cast
import subprocess
import termios
import sys

from tqdm import tqdm

from .status import NoOpen, ToOpen, Broken, Exists, Missing, Tiny, PointFive

if TYPE_CHECKING:
    from .state import State, URL


_order = (Broken, Exists, Missing, PointFive, Tiny)


def _fmt(x: URL) -> str:
    if (exc := getattr(x.status, "exc", None)) is not None:
        return f"{x.url}: {exc}"
    return x.url


def print_no_open_failures(state: State) -> None:
    no_open: tuple[URL, ...] = state.get(NoOpen)
    if no_open:
        print(f"{'*'*70}\n*{'Errors'.center(68)}*\n{'*'*70}\n")
        for type_ in {cast(NoOpen, i.status).__class__ for i in no_open}:  # Set de-dups
            got = sorted(state.get(type_), key=lambda x: x.url)
            print(f"{type_.kind()}\n\t" + "\n\t".join(_fmt(i) for i in got) + "\n")
        print(f"{'*' * 70}\n*{'Done'.center(68)}*\n{'*' * 70}\n")


def print_open_failures(state: State, no_prompt: bool, skip_tiny: bool, skip_point_five: bool, opener: str) -> None:
    to_open: tuple[URL, ...] = state.get(ToOpen)
    if not to_open:
        return
    if not no_prompt:
        termios.tcflush(sys.stdin, termios.TCIFLUSH)
        _ = input("Testing complete. Hit enter to open websites.")
    classes = {cast(ToOpen, i.status).__class__ for i in to_open}  # Set de-dups
    for type_ in sorted(classes, key=lambda t: _order.index(t) if t in _order else len(_order)):
        if skip_tiny and type_ is Tiny:
            print("Skipping tiny\n")
            continue
        if skip_point_five and type_ is PointFive:
            print("Skipping .5\n")
            continue
        got = state.get(type_)
        print(f"{type_.kind()}\n\t" + "\n\t".join(_fmt(i) for i in got))
        for url in tqdm(got, dynamic_ncols=True, leave=False):
            subprocess.check_call([opener, url.url], stdout=subprocess.DEVNULL)
        print("")


def results(state: State, no_prompt: bool, skip_tiny: bool, skip_point_five: bool, opener: str) -> None:
    print_no_open_failures(state)
    print_open_failures(state, no_prompt, skip_tiny, skip_point_five, opener)
    print("Done!")
